@@ -1,7 +1,7 @@
 /* Determine whether string value is affirmation or negative response
    according to current locale's data.
 
-   Copyright (C) 1996, 1998, 2000, 2002-2003, 2006-2023 Free Software
+   Copyright (C) 1996, 1998, 2000, 2002-2003, 2006-2025 Free Software
    Foundation, Inc.
 
    This program is free software: you can redistribute it and/or modify
@@ -33,7 +33,7 @@
 # endif
 # include <regex.h>
 # include "gettext.h"
-# define _(msgid) gettext (msgid)
+# define _(msgid) dgettext ("gnulib", msgid)
 # define N_(msgid) gettext_noop (msgid)
 
 # if HAVE_LANGINFO_YESEXPR
