@@ -1,5 +1,5 @@
 /* This file is part of GNU paxutils
-   Copyright (C) 2005, 2007 Free Software Foundation, Inc.
+   Copyright (C) 2005, 2007, 2010, 2023-2025 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
@@ -12,8 +12,7 @@
    Public License for more details.
 
    You should have received a copy of the GNU General Public License along
-   with this program; if not, write to the Free Software Foundation, Inc.,
-   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */
+   with this program.  If not, see <http://www.gnu.org/licenses/>. */
 
 #include <system.h>
 #include <hash.h>
@@ -38,28 +37,19 @@ hash_string_compare (void const *name1, void const *name2)
 
 /* Return zero if TABLE contains a LEN-character long prefix of STRING,
    otherwise, insert a newly allocated copy of this prefix to TABLE and
-   return 1.  If RETURN_PREFIX is not NULL, point it to the allocated
+   return true.  If RETURN_PREFIX is nonnull, point it to the allocated
    copy. */
 static bool
 hash_string_insert_prefix (Hash_table **table, char const *string, size_t len,
 			   const char **return_prefix)
 {
   Hash_table *t = *table;
-  char *s;
+  char *s = ximemdup0 (string, len);
   char *e;
 
-  if (len)
-    {
-      s = xmalloc (len + 1);
-      memcpy (s, string, len);
-      s[len] = 0;
-    }
-  else
-    s = xstrdup (string);
-  
   if (! ((t
-	  || (*table = t = hash_initialize (0, 0, hash_string_hasher,
-					    hash_string_compare, 0)))
+	  || (*table = t = hash_initialize (0, nullptr, hash_string_hasher,
+					    hash_string_compare, nullptr)))
 	 && (e = hash_insert (t, s))))
     xalloc_die ();
 
@@ -67,30 +57,15 @@ hash_string_insert_prefix (Hash_table **table, char const *string, size_t len,
     {
       if (return_prefix)
 	*return_prefix = s;
-      return 1;
+      return true;
     }
   else
     {
       free (s);
-      return 0;
+      return false;
     }
 }
 
-/* Return zero if TABLE contains a copy of STRING; otherwise, insert a
-   copy of STRING to TABLE and return 1.  */
-bool
-hash_string_insert (Hash_table **table, char const *string)
-{
-  return hash_string_insert_prefix (table, string, 0, NULL);
-}
-
-/* Return 1 if TABLE contains STRING.  */
-bool
-hash_string_lookup (Hash_table const *table, char const *string)
-{
-  return table && hash_lookup (table, string);
-}
-
 
 static Hash_table *prefix_table[2];
 
@@ -105,57 +80,52 @@ removed_prefixes_p (void)
          || (prefix_table[1] && hash_get_n_entries (prefix_table[1]) != 0);
 }
 
-/* Return a safer suffix of FILE_NAME, or "." if it has no safer
-   suffix.  Check for fully specified file names and other atrocities.
-   Warn the user if we do not return NAME.  If LINK_TARGET is 1,
+/* Return a safer suffix of FILE_NAME, or "." if it has no safer suffix.
+   Skip any sequence of prefixes each of which would cause
+   the file name to escape the working directory on this platform.
+   Warn the user if we do not return NAME.  If LINK_TARGET,
    FILE_NAME is the target of a hard link, not a member name.
-   If ABSOLUTE_NAMES is 0, strip filesystem prefix from the file name. */
+   However, if ABSOLUTE_NAMES, do not skip prefixes, but instead
+   return FILE_NAME if nonempty, "." otherwise.  */
 
 char *
 safer_name_suffix (char const *file_name, bool link_target,
 		   bool absolute_names)
 {
-  char const *p;
+  char const *p = file_name;
 
-  if (absolute_names)
-    p = file_name;
-  else
+  if (!absolute_names)
     {
-      /* Skip file system prefixes, leading file name components that contain
-	 "..", and leading slashes.  */
-
-      size_t prefix_len = FILE_SYSTEM_PREFIX_LEN (file_name);
-
-      for (p = file_name + prefix_len; *p; )
+      /* Skip any sequences of prefixes each of which would cause the
+	 resulting file name to escape the working directory on this platform.
+         The resulting file name is relative, not absolute.  */
+      for (;;)
 	{
-          if (p[0] == '.' && p[1] == '.' && (ISSLASH (p[2]) || !p[2]))
-	    prefix_len = p + 2 - file_name;
-
-	  do
+	  if (ISSLASH (*p))
+	    p++;
+	  else if (p[0] == '.' && p[1] == '.' && (ISSLASH (p[2]) || !p[2]))
+	    p += 2;
+	  else
 	    {
-	      char c = *p++;
-	      if (ISSLASH (c))
+	      int prefix_len = FILE_SYSTEM_PREFIX_LEN (p);
+	      if (prefix_len == 0)
 		break;
+	      p += prefix_len;
 	    }
-	  while (*p);
 	}
 
-      for (p = file_name + prefix_len; ISSLASH (*p); p++)
-	continue;
-      prefix_len = p - file_name;
-
-      if (prefix_len)
+      if (p != file_name)
 	{
 	  const char *prefix;
 	  if (hash_string_insert_prefix (&prefix_table[link_target], file_name,
-					 prefix_len, &prefix))
+					 p - file_name, &prefix))
 	    {
 	      static char const *const diagnostic[] =
 	      {
 		N_("Removing leading `%s' from member names"),
 		N_("Removing leading `%s' from hard link targets")
 	      };
-	      WARN ((0, 0, _(diagnostic[link_target]), prefix));
+	      paxwarn (0, _(diagnostic[link_target]), prefix);
 	    }
 	}
     }
@@ -169,7 +139,7 @@ safer_name_suffix (char const *file_name, bool link_target,
 	    N_("Substituting `.' for empty member name"),
 	    N_("Substituting `.' for empty hard link target")
 	  };
-	  WARN ((0, 0, "%s", _(diagnostic[link_target])));
+	  paxwarn (0, "%s", _(diagnostic[link_target]));
 	}
 
       p = ".";
