@@ -1,6 +1,6 @@
 /* Diff files from a tar archive.
 
-   Copyright 1988-2019 Free Software Foundation, Inc.
+   Copyright 1988-2025 Free Software Foundation, Inc.
 
    This file is part of GNU tar.
 
@@ -20,15 +20,21 @@
    Written by John Gilmore, on 1987-04-30.  */
 
 #include <system.h>
-#include <system-ioctl.h>
 
 #if HAVE_LINUX_FD_H
 # include <linux/fd.h>
 #endif
 
+#if HAVE_SYS_MTIO_H
+# include <sys/ioctl.h>
+# include <sys/mtio.h>
+#endif
+
 #include "common.h"
+#include <alignalloc.h>
 #include <quotearg.h>
 #include <rmt.h>
+#include <same-inode.h>
 #include <stdarg.h>
 
 /* Nonzero if we are verifying at the moment.  */
@@ -44,8 +50,7 @@ static char *diff_buffer;
 void
 diff_init (void)
 {
-  void *ptr;
-  diff_buffer = page_aligned_alloc (&ptr, record_size);
+  diff_buffer = xalignalloc (getpagesize (), record_size);
   if (listed_incremental_option)
     read_directory_file ();
 }
@@ -72,21 +77,20 @@ report_difference (struct tar_stat_info *st, const char *fmt, ...)
 }
 
 /* Take a buffer returned by read_and_process and do nothing with it.  */
-static int
-process_noop (size_t size __attribute__ ((unused)),
-	      char *data __attribute__ ((unused)))
+static bool
+process_noop (MAYBE_UNUSED idx_t size, MAYBE_UNUSED char *data)
 {
-  return 1;
+  return true;
 }
 
-static int
-process_rawdata (size_t bytes, char *buffer)
+static bool
+process_rawdata (idx_t bytes, char *buffer)
 {
-  size_t status = blocking_read (diff_handle, diff_buffer, bytes);
+  idx_t status = blocking_read (diff_handle, diff_buffer, bytes);
 
-  if (status != bytes)
+  if (status < bytes)
     {
-      if (status == SAFE_READ_ERROR)
+      if (errno)
 	{
 	  read_error (current_stat_info.file_name);
 	  report_difference (&current_stat_info, NULL);
@@ -94,53 +98,48 @@ process_rawdata (size_t bytes, char *buffer)
       else
 	{
 	  report_difference (&current_stat_info,
-			     ngettext ("Could only read %lu of %lu byte",
-				       "Could only read %lu of %lu bytes",
+			     ngettext ("Could read only %td of %td byte",
+				       "Could read only %td of %td bytes",
 				       bytes),
-			     (unsigned long) status, (unsigned long) bytes);
+			     status, bytes);
 	}
-      return 0;
+      return false;
     }
 
-  if (memcmp (buffer, diff_buffer, bytes))
+  if (memcmp (buffer, diff_buffer, bytes) != 0)
     {
       report_difference (&current_stat_info, _("Contents differ"));
-      return 0;
+      return false;
     }
 
-  return 1;
+  return true;
 }
 
-/* Some other routine wants SIZE bytes in the archive.  For each chunk
-   of the archive, call PROCESSOR with the size of the chunk, and the
-   address of the chunk it can work with.  The PROCESSOR should return
-   nonzero for success.  Once it returns error, continue skipping
-   without calling PROCESSOR anymore.  */
+/* Some other routine wants ST->stat.st_size bytes in the archive.
+   For each chunk of the archive, call PROCESSOR with the size of the chunk,
+   and the address of the chunk it can work with.  PROCESSOR should return
+   true for success.  Once it fails, continue skipping without calling
+   PROCESSOR anymore.  */
 
 static void
-read_and_process (struct tar_stat_info *st, int (*processor) (size_t, char *))
+read_and_process (struct tar_stat_info *st, bool (*processor) (idx_t, char *))
 {
-  union block *data_block;
-  size_t data_size;
-  off_t size = st->stat.st_size;
-
   mv_begin_read (st);
-  while (size)
+  for (off_t size = st->stat.st_size; size; )
     {
-      data_block = find_next_block ();
+      union block *data_block = find_next_block ();
       if (! data_block)
 	{
-	  ERROR ((0, 0, _("Unexpected EOF in archive")));
+	  paxerror (0, _("Unexpected EOF in archive"));
 	  return;
 	}
 
-      data_size = available_space_after (data_block);
+      idx_t data_size = available_space_after (data_block);
       if (data_size > size)
 	data_size = size;
-      if (!(*processor) (data_size, data_block->buffer))
+      if (!processor (data_size, charptr (data_block)))
 	processor = process_noop;
-      set_next_block_after ((union block *)
-			    (data_block->buffer + data_size - 1));
+      set_next_block_after (charptr (data_block) + data_size - 1);
       size -= data_size;
       mv_size_left (size);
     }
@@ -149,23 +148,18 @@ read_and_process (struct tar_stat_info *st, int (*processor) (size_t, char *))
 
 /* Call either stat or lstat over STAT_DATA, depending on
    --dereference (-h), for a file which should exist.  Diagnose any
-   problem.  Return nonzero for success, zero otherwise.  */
-static int
+   problem.  Return true for success, false otherwise.  */
+static bool
 get_stat_data (char const *file_name, struct stat *stat_data)
 {
-  int status = deref_stat (file_name, stat_data);
-
-  if (status != 0)
+  if (deref_stat (file_name, stat_data) < 0)
     {
-      if (errno == ENOENT)
-	stat_warn (file_name);
-      else
-	stat_error (file_name);
+      (errno == ENOENT ? stat_warn : stat_error) (file_name);
       report_difference (&current_stat_info, NULL);
-      return 0;
+      return false;
     }
 
-  return 1;
+  return true;
 }
 
 
@@ -229,8 +223,6 @@ diff_file (void)
 	    }
 	  else
 	    {
-	      int status;
-
 	      if (current_stat_info.is_sparse)
 		sparse_diff_file (diff_handle, &current_stat_info);
 	      else
@@ -241,12 +233,11 @@ diff_file (void)
 		{
 		  struct timespec atime = get_stat_atime (&stat_data);
 		  if (set_file_atime (diff_handle, chdir_fd, file_name, atime)
-		      != 0)
+		      < 0)
 		    utime_error (file_name);
 		}
 
-	      status = close (diff_handle);
-	      if (status != 0)
+	      if (close (diff_handle) < 0)
 		close_error (file_name);
 	    }
 	}
@@ -261,19 +252,18 @@ diff_link (void)
 
   if (get_stat_data (current_stat_info.file_name, &file_data)
       && get_stat_data (current_stat_info.link_name, &link_data)
-      && !sys_compare_links (&file_data, &link_data))
+      && !psame_inode (&file_data, &link_data))
     report_difference (&current_stat_info,
 		       _("Not linked to %s"),
 		       quote_n_colon (QUOTE_ARG,
 				      current_stat_info.link_name));
 }
 
-#ifdef HAVE_READLINK
 static void
 diff_symlink (void)
 {
   char buf[1024];
-  size_t len = strlen (current_stat_info.link_name);
+  idx_t len = strlen (current_stat_info.link_name);
   char *linkbuf = len < sizeof buf ? buf : xmalloc (len + 1);
 
   ssize_t status = readlinkat (chdir_fd, current_stat_info.file_name,
@@ -294,7 +284,6 @@ diff_symlink (void)
   if (linkbuf != buf)
     free (linkbuf);
 }
-#endif
 
 static void
 diff_special (void)
@@ -330,37 +319,42 @@ diff_special (void)
     report_difference (&current_stat_info, _("Mode differs"));
 }
 
-static int
+/* Return zero if and only if A and B should be considered equal.
+   for the purposes of dump directory comparison.  */
+static char
 dumpdir_cmp (const char *a, const char *b)
 {
-  size_t len;
-
   while (*a)
     switch (*a)
       {
       case 'Y':
       case 'N':
-	if (!strchr ("YN", *b))
+	/* If the null-terminated strings A and B are equal, other
+	   than possibly A's first byte being 'Y' where B's is 'N' or
+	   vice versa, advance A and B past the strings.
+	   Otherwise, return 1.  */
+	if (! (*b == 'Y' || *b == 'N'))
 	  return 1;
-	if (strcmp(a + 1, b + 1))
-	  return 1;
-	len = strlen (a) + 1;
-	a += len;
-	b += len;
-	break;
-
+	a++, b++;
+	FALLTHROUGH;
       case 'D':
-	if (strcmp(a, b))
+	/* If the null-terminated strings A and B are equal, advance A
+	   and B past them.  Otherwise, return 1.  */
+	while (*a)
+	  if (*a++ != *b++)
+	    return 1;
+	if (*b)
 	  return 1;
-	len = strlen (a) + 1;
-	a += len;
-	b += len;
+	a++, b++;
 	break;
 
       case 'R':
       case 'T':
       case 'X':
 	return *b;
+
+      default:
+	unreachable ();
       }
   return *b;
 }
@@ -376,7 +370,7 @@ diff_dumpdir (struct tar_stat_info *dir)
       int fd = subfile_open (dir->parent, dir->orig_file_name, open_read_flags);
       if (fd < 0)
 	diag = open_diag;
-      else if (fstat (fd, &dir->stat))
+      else if (fstat (fd, &dir->stat) < 0)
         {
 	  diag = stat_diag;
           close (fd);
@@ -393,7 +387,7 @@ diff_dumpdir (struct tar_stat_info *dir)
 
   if (dumpdir_buffer)
     {
-      if (dumpdir_cmp (dir->dumpdir, dumpdir_buffer))
+      if (dumpdir_cmp (dir->dumpdir, dumpdir_buffer) != 0)
 	report_difference (dir, _("Contents differ"));
     }
   else
@@ -403,16 +397,13 @@ diff_dumpdir (struct tar_stat_info *dir)
 static void
 diff_multivol (void)
 {
-  struct stat stat_data;
-  int fd, status;
-  off_t offset;
-
   if (current_stat_info.had_trailing_slash)
     {
       diff_dir ();
       return;
     }
 
+  struct stat stat_data;
   if (!get_stat_data (current_stat_info.file_name, &stat_data))
     return;
 
@@ -423,10 +414,11 @@ diff_multivol (void)
       return;
     }
 
-  offset = OFF_FROM_HEADER (current_header->oldgnu_header.offset);
+  off_t offset = OFF_FROM_HEADER (current_header->oldgnu_header.offset);
+  off_t file_size;
   if (offset < 0
-      || INT_ADD_OVERFLOW (current_stat_info.stat.st_size, offset)
-      || stat_data.st_size != current_stat_info.stat.st_size + offset)
+      || ckd_add (&file_size, current_stat_info.stat.st_size, offset)
+      || stat_data.st_size != file_size)
     {
       report_difference (&current_stat_info, _("Size differs"));
       skip_member ();
@@ -434,7 +426,7 @@ diff_multivol (void)
     }
 
 
-  fd = openat (chdir_fd, current_stat_info.file_name, open_read_flags);
+  int fd = openat (chdir_fd, current_stat_info.file_name, open_read_flags);
 
   if (fd < 0)
     {
@@ -452,8 +444,7 @@ diff_multivol (void)
   else
     read_and_process (&current_stat_info, process_rawdata);
 
-  status = close (fd);
-  if (status != 0)
+  if (close (fd) < 0)
     close_error (current_stat_info.file_name);
 }
 
@@ -476,9 +467,9 @@ diff_archive (void)
   switch (current_header->header.typeflag)
     {
     default:
-      ERROR ((0, 0, _("%s: Unknown file type '%c', diffed as normal file"),
-	      quotearg_colon (current_stat_info.file_name),
-	      current_header->header.typeflag));
+      paxerror (0, _("%s: Unknown file type '%c', diffed as normal file"),
+		quotearg_colon (current_stat_info.file_name),
+		current_header->header.typeflag);
       FALLTHROUGH;
     case AREGTYPE:
     case REGTYPE:
@@ -497,11 +488,9 @@ diff_archive (void)
       diff_link ();
       break;
 
-#ifdef HAVE_READLINK
     case SYMTYPE:
       diff_symlink ();
       break;
-#endif
 
     case CHRTYPE:
     case BLKTYPE:
@@ -527,22 +516,20 @@ diff_archive (void)
 void
 verify_volume (void)
 {
-  int may_fail = 0;
+  bool may_fail = false;
   if (removed_prefixes_p ())
     {
-      WARN((0, 0,
-	    _("Archive contains file names with leading prefixes removed.")));
-      may_fail = 1;
+      paxwarn (0,
+	       _("Archive contains file names with leading prefixes removed."));
+      may_fail = true;
     }
   if (transform_program_p ())
     {
-      WARN((0, 0,
-	    _("Archive contains transformed file names.")));
-      may_fail = 1;
+      paxwarn (0, _("Archive contains transformed file names."));
+      may_fail = true;
     }
   if (may_fail)
-    WARN((0, 0,
-	  _("Verification may fail to locate original files.")));
+    paxwarn (0, _("Verification may fail to locate original files."));
 
   clear_directory_table ();
 
@@ -566,31 +553,12 @@ verify_volume (void)
   ioctl (archive, FDFLUSH);
 #endif
 
-#ifdef MTIOCTOP
-  {
-    struct mtop operation;
-    int status;
-
-    operation.mt_op = MTBSF;
-    operation.mt_count = 1;
-    if (status = rmtioctl (archive, MTIOCTOP, (char *) &operation), status < 0)
-      {
-	if (errno != EIO
-	    || (status = rmtioctl (archive, MTIOCTOP, (char *) &operation),
-		status < 0))
-	  {
-#endif
-	    if (rmtlseek (archive, (off_t) 0, SEEK_SET) != 0)
-	      {
-		/* Lseek failed.  Try a different method.  */
-		seek_warn (archive_name_array[0]);
-		return;
-	      }
-#ifdef MTIOCTOP
-	  }
-      }
-  }
-#endif
+  if (!mtioseek (true, -1) && rmtlseek (archive, 0, SEEK_SET) < 0)
+    {
+      /* Lseek failed.  Try a different method.  */
+      seek_warn (archive_name_array[0]);
+      return;
+    }
 
   access_mode = ACCESS_READ;
   now_verifying = 1;
@@ -604,7 +572,7 @@ verify_volume (void)
 
       if (status == HEADER_FAILURE)
 	{
-	  int counter = 0;
+	  intmax_t counter = 0;
 
 	  do
 	    {
@@ -615,10 +583,11 @@ verify_volume (void)
 	    }
 	  while (status == HEADER_FAILURE);
 
-	  ERROR ((0, 0,
-		  ngettext ("VERIFY FAILURE: %d invalid header detected",
-			    "VERIFY FAILURE: %d invalid headers detected",
-			    counter), counter));
+	  paxerror (0,
+		    ngettext ("VERIFY FAILURE: %jd invalid header detected",
+			      "VERIFY FAILURE: %jd invalid headers detected",
+			      counter),
+		    counter);
 	}
       if (status == HEADER_END_OF_FILE)
 	break;
@@ -627,20 +596,18 @@ verify_volume (void)
 	  set_next_block_after (current_header);
           if (!ignore_zeros_option)
             {
-	      char buf[UINTMAX_STRSIZE_BOUND];
-
 	      status = read_header (&current_header, &current_stat_info,
 	                            read_header_auto);
 	      if (status == HEADER_ZERO_BLOCK)
 	        break;
-	      WARNOPT (WARN_ALONE_ZERO_BLOCK,
-		       (0, 0, _("A lone zero block at %s"),
-			STRINGIFY_BIGINT (current_block_ordinal (), buf)));
+	      warnopt (WARN_ALONE_ZERO_BLOCK,
+		       0, _("A lone zero block at %jd"),
+		       intmax (current_block_ordinal ()));
             }
 	  continue;
 	}
 
-      decode_header (current_header, &current_stat_info, &current_format, 1);
+      decode_header (current_header, &current_stat_info, &current_format, true);
       diff_archive ();
       tar_stat_destroy (&current_stat_info);
     }
